// ***************************************************************************
// ***************************************************************************
// ***************************************************************************
// ***************************************************************************

#include <stdio.h>
#include "platform.h"
#include "xbasic_types.h"
#include "xstatus.h"
#include "xil_io.h"
#include "xparameters.h"
#include "xuartlite_l.h"

#define JESD_BASEADDR    XPAR_AXI_JESD204B_RX4_0_BASEADDR
#define DMA0_BASEADDR    XPAR_AXI_DMA_0_BASEADDR
#define DMA1_BASEADDR    XPAR_AXI_DMA_1_BASEADDR
#define VDMA0_BASEADDR   XPAR_AXI_VDMA_0_BASEADDR
#define VDMA1_BASEADDR   XPAR_AXI_VDMA_1_BASEADDR
#define VDMA2_BASEADDR   XPAR_AXI_VDMA_2_BASEADDR
#define DDR_BASEADDR     XPAR_DDR3_SDRAM_S_AXI_BASEADDR
#define UART_BASEADDR    XPAR_RS232_UART_1_BASEADDR
#define SPI_BASEADDR     XPAR_AXI_SPI_0_BASEADDR
#define IIC_BASEADDR     XPAR_AXI_IIC_0_BASEADDR
#define GPIOLCD_BASEADDR XPAR_AXI_GPIO_0_BASEADDR
#define CF_ADC0_BASEADDR XPAR_AXI_AD9250_0_BASEADDR
#define CF_ADC1_BASEADDR XPAR_AXI_AD9250_1_BASEADDR
#define CF_DAC0_BASEADDR XPAR_AXI_AD9129_0_BASEADDR
#define CF_DAC1_BASEADDR XPAR_AXI_AD9129_1_BASEADDR
#define CF_CGEN_BASEADDR XPAR_AXI_CLKGEN_0_BASEADDR
#define CF_HDMI_BASEADDR XPAR_AXI_HDMI_TX_16B_0_BASEADDR

#define FMC176_CPLD      0x00
#define FMC176_AD9517    0x84
#define FMC176_AD9250_0  0x80
#define FMC176_AD9250_1  0x81
#define FMC176_AD9129_0  0x82
#define FMC176_AD9129_1  0x83

#define H_STRIDE(m)      (((m) == 1) ? 1280 : 1920)
#define H_COUNT(m)       (((m) == 1) ? 1650 : 2200)
#define H_ACTIVE(m)      (((m) == 1) ? 1280 : 1920)
#define H_WIDTH(m)       (((m) == 1) ?   40 :   44)
#define H_FP(m)          (((m) == 1) ?  110 :   88)
#define H_BP(m)          (((m) == 1) ?  220 :  148)
#define V_COUNT(m)       (((m) == 1) ?  750 : 1125)
#define V_ACTIVE(m)      (((m) == 1) ?  720 : 1080)
#define V_WIDTH(m)       (((m) == 1) ?    5 :    5)
#define V_FP(m)          (((m) == 1) ?    5 :    4)
#define V_BP(m)          (((m) == 1) ?   20 :   36)

#define H_DE_MIN(m)      (H_WIDTH(m) + H_BP(m))
#define H_DE_MAX(m)      (H_WIDTH(m) + H_BP(m) + H_ACTIVE(m))
#define V_DE_MIN(m)      (V_WIDTH(m) + V_BP(m))
#define V_DE_MAX(m)      (V_WIDTH(m) + V_BP(m) + V_ACTIVE(m))
#define VIDEO_LENGTH(m)  (H_ACTIVE(m) * V_ACTIVE(m))
#define VIDEO_BASEADDR   DDR_BASEADDR + 0x2000000

// ***************************************************************************
// ***************************************************************************

extern char inbyte(void);
void xil_printf(const char *ctrl1, ...);
void microblaze_flush_dcache(void);
void microblaze_invalidate_dcache(void);

void delay_ms(u32 ms_count) {
  u32 count;
  for (count = 0; count < ((ms_count * 100000) + 1); count++) {
    asm("nop");
  }
}

u32 user_exit(void) {
  while (!XUartLite_IsReceiveEmpty(UART_BASEADDR)) {
    if (XUartLite_RecvByte(UART_BASEADDR) == 'q') {
      return(1);
    }
  }
  return(0);
}

// ***************************************************************************
// ***************************************************************************

void lcd_write_1(u32 data) {
  u32 count;
  Xil_Out32(GPIOLCD_BASEADDR, (0x00 | (data&0x3f)));
  Xil_Out32(GPIOLCD_BASEADDR, (0x00 | (data&0x3f)));
  for (count = 0; count < 13; count++) {
    asm("nop");
  }
  Xil_Out32(GPIOLCD_BASEADDR, (0x40 | (data&0x3f)));
  Xil_Out32(GPIOLCD_BASEADDR, (0x40 | (data&0x3f)));
  for (count = 0; count < 26; count++) {
    asm("nop");
  }
  Xil_Out32(GPIOLCD_BASEADDR, (0x00 | (data&0x3f)));
  Xil_Out32(GPIOLCD_BASEADDR, (0x00 | (data&0x3f)));
  for (count = 0; count < 13; count++) {
    asm("nop");
  }
}

void lcd_write(u32 cntrl, u32 data) {
  u32 count;
  lcd_write_1((((cntrl<<4)&0x30) | ((data>>4)&0xf)));
  for (count = 0; count < 100; count++) {
    asm("nop");
  }
  lcd_write_1((((cntrl<<4)&0x30) | ((data>>0)&0xf)));
  for (count = 0; count < 4000; count++) {
    asm("nop");
  }
}

void lcd_display(u8 no_of_dac_channels, u8 no_of_adc_channels) {

  Xil_Out32(GPIOLCD_BASEADDR, 0x00);
  Xil_Out32((GPIOLCD_BASEADDR + 0x04), 0x00); // outputs

  delay_ms(16);
  lcd_write_1(0x3);
  delay_ms(5);
  lcd_write_1(0x3);
  delay_ms(1);
  lcd_write_1(0x3);
  delay_ms(1);
  lcd_write_1(0x2);
  delay_ms(1);

  lcd_write(0x0, 0x28);
  lcd_write(0x0, 0x06);
  lcd_write(0x0, 0x0c);
  lcd_write(0x0, 0x01);
  delay_ms(2);

  lcd_write(0x0, 0x02);
  lcd_write(0x2, 0x41);
  lcd_write(0x2, 0x6e);
  lcd_write(0x2, 0x61);
  lcd_write(0x2, 0x6c);
  lcd_write(0x2, 0x6f);
  lcd_write(0x2, 0x67);
  lcd_write(0x2, 0x20);
  lcd_write(0x2, 0x44);
  lcd_write(0x2, 0x65);
  lcd_write(0x2, 0x76);
  lcd_write(0x2, 0x69);
  lcd_write(0x2, 0x63);
  lcd_write(0x2, 0x65);
  lcd_write(0x2, 0x73);
  lcd_write(0x2, 0x20);

  lcd_write(0x0, 0xa8);
  lcd_write(0x2, (0x30 + no_of_adc_channels));
  lcd_write(0x2, 0x41);
  lcd_write(0x2, 0x44);
  lcd_write(0x2, 0x39);
  lcd_write(0x2, 0x32);
  lcd_write(0x2, 0x35);
  lcd_write(0x2, 0x30);
  lcd_write(0x2, 0x2d);
  lcd_write(0x2, (0x30 + no_of_dac_channels));
  lcd_write(0x2, 0x41);
  lcd_write(0x2, 0x44);
  lcd_write(0x2, 0x39);
  lcd_write(0x2, 0x31);
  lcd_write(0x2, 0x32);
  lcd_write(0x2, 0x39);
  lcd_write(0x2, 0x20);

  xil_printf("Analog Devices, %dAD9250-%dAD9129\n\r", no_of_adc_channels, no_of_dac_channels);
}

// ***************************************************************************
// ***************************************************************************

void fmc176_spi_write(u32 sel, u32 addr, u32 data) {
  Xil_Out32((SPI_BASEADDR + 0x70), 0x001); // chip select
  Xil_Out32((SPI_BASEADDR + 0x60), 0x1e6); // control register (rst fifos)
  Xil_Out32((SPI_BASEADDR + 0x70), 0x000); // chip select
  Xil_Out32((SPI_BASEADDR + 0x68), (sel & 0xff)); // mosi data
  Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
  while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
  Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
  Xil_In32(SPI_BASEADDR + 0x6c);
  if ((sel == FMC176_AD9517) || (sel == FMC176_AD9250_0) || (sel == FMC176_AD9250_1)) {
    Xil_Out32((SPI_BASEADDR + 0x68), ((addr>>8) & 0xff)); // mosi data
    Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
    while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
    Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
    Xil_In32(SPI_BASEADDR + 0x6c);
  }
  Xil_Out32((SPI_BASEADDR + 0x68), (addr & 0xff)); // mosi data
  Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
  while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
  Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
  Xil_In32(SPI_BASEADDR + 0x6c);
  Xil_Out32((SPI_BASEADDR + 0x68), data); // mosi data
  Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
  while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
  Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
  Xil_In32(SPI_BASEADDR + 0x6c);
  Xil_Out32((SPI_BASEADDR + 0x70), 0x001); // chip select
  Xil_Out32((SPI_BASEADDR + 0x60), 0x180); // control register
  delay_ms(1);
}

u32 fmc176_spi_read(u32 sel, u32 addr, u32 display) {
  u32 rdata;
  rdata = 0x80;
  Xil_Out32((SPI_BASEADDR + 0x70), 0x001); // chip select
  Xil_Out32((SPI_BASEADDR + 0x60), 0x1e6); // control register (rst fifos)
  Xil_Out32((SPI_BASEADDR + 0x70), 0x000); // chip select
  Xil_Out32((SPI_BASEADDR + 0x68), (sel & 0xff)); // mosi data
  Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
  while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
  Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
  Xil_In32(SPI_BASEADDR + 0x6c);
  if ((sel == FMC176_AD9517) || (sel == FMC176_AD9250_0) || (sel == FMC176_AD9250_1)) {
    Xil_Out32((SPI_BASEADDR + 0x68), (rdata | ((addr>>8) & 0xff))); // mosi data
    Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
    while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
    Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
    Xil_In32(SPI_BASEADDR + 0x6c);
    rdata = 0x00;
  }
  Xil_Out32((SPI_BASEADDR + 0x68), (rdata | (addr & 0xff))); // mosi data
  Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
  while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
  Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
  Xil_In32(SPI_BASEADDR + 0x6c);
  Xil_Out32((SPI_BASEADDR + 0x68), 0x000); // mosi data
  Xil_Out32((SPI_BASEADDR + 0x60), 0x086); // control register (rst fifos)
  while ((Xil_In32((SPI_BASEADDR + 0x64)) & 0x4) == 0x0) {}
  Xil_Out32((SPI_BASEADDR + 0x60), 0x186); // control register (rst fifos)
  rdata = Xil_In32(SPI_BASEADDR + 0x6c);
  Xil_Out32((SPI_BASEADDR + 0x70), 0x001); // chip select
  Xil_Out32((SPI_BASEADDR + 0x60), 0x180); // control register
  delay_ms(1);
  if (display == 1) {
    if (sel == FMC176_CPLD)      xil_printf("FMC176_CPLD");
    if (sel == FMC176_AD9517)    xil_printf("FMC176_AD9517");
    if (sel == FMC176_AD9250_0)  xil_printf("FMC176_AD9250_0");
    if (sel == FMC176_AD9250_1)  xil_printf("FMC176_AD9250_1");
    if (sel == FMC176_AD9129_0)  xil_printf("FMC176_AD9129_0");
    if (sel == FMC176_AD9129_1)  xil_printf("FMC176_AD9129_1");
    xil_printf("[%04x]: %02x\n\r", addr, rdata);
  }
  return(rdata);
}

void fmc176_spi_verify(u32 sel, u32 addr, u32 data) {
  u32 rdata;
  rdata = fmc176_spi_read(sel, addr, 0);
  if (rdata != data) {
    if (sel == FMC176_CPLD)      xil_printf("FMC176_CPLD");
    if (sel == FMC176_AD9517)    xil_printf("FMC176_AD9517");
    if (sel == FMC176_AD9250_0)  xil_printf("FMC176_AD9250_0");
    if (sel == FMC176_AD9250_1)  xil_printf("FMC176_AD9250_1");
    if (sel == FMC176_AD9129_0)  xil_printf("FMC176_AD9129_0");
    if (sel == FMC176_AD9129_1)  xil_printf("FMC176_AD9129_1");
    xil_printf("[%04x]: rcv(%02x), exp(%02x) !!!\n\r", addr, rdata, data);
  }
  return;
}

// ***************************************************************************
// ***************************************************************************

u32 ad9250_sel(u32 sel) {
  if (sel == 0) return(FMC176_AD9250_0);
  if (sel == 1) return(FMC176_AD9250_1);
  return(-1);
}

u32 ad9129_sel(u32 sel) {
  if (sel == 0) return(FMC176_AD9129_0);
  if (sel == 1) return(FMC176_AD9129_1);
  return(-1);
}

// ***************************************************************************
// ***************************************************************************

u32 ad9517_setup() {

  // reference clock is 30.72MHz. vco is at 2.4576GHz (ref1/not-lvds)
  // r divider = 1, n divider 80 (a = 0, b = 5), prescaler- 16/17 DM Mode
  // adc1 (out0) and adc0 (out1) are at 245.76MHz
  // dac1 (out3) and dac0 (out2) are at 2.4576GHz (vco clock)
  // fpga clock (out4) is at 245.76MHz
  // gtx clock (out5) is at 245.76MHz
  // out clock (out6) is at 245.76MHz

  fmc176_spi_write(FMC176_AD9517, 0x0010, 0x7c);
  fmc176_spi_write(FMC176_AD9517, 0x0014, 0x05);
  fmc176_spi_write(FMC176_AD9517, 0x0016, 0x05);
  fmc176_spi_write(FMC176_AD9517, 0x001c, 0x02);
  fmc176_spi_write(FMC176_AD9517, 0x00f1, 0x08);
  fmc176_spi_write(FMC176_AD9517, 0x00f5, 0x08);
  fmc176_spi_write(FMC176_AD9517, 0x0140, 0x42);
  fmc176_spi_write(FMC176_AD9517, 0x0141, 0x42);
  fmc176_spi_write(FMC176_AD9517, 0x0142, 0x42);
  fmc176_spi_write(FMC176_AD9517, 0x0143, 0x43);
  fmc176_spi_write(FMC176_AD9517, 0x0190, 0x21);
  fmc176_spi_write(FMC176_AD9517, 0x0191, 0x00);
  fmc176_spi_write(FMC176_AD9517, 0x0197, 0x80);
  fmc176_spi_write(FMC176_AD9517, 0x0198, 0x02);
  fmc176_spi_write(FMC176_AD9517, 0x0199, 0x21);
  fmc176_spi_write(FMC176_AD9517, 0x019c, 0x20);
  fmc176_spi_write(FMC176_AD9517, 0x019e, 0x21);
  fmc176_spi_write(FMC176_AD9517, 0x01a0, 0x11);
  fmc176_spi_write(FMC176_AD9517, 0x01a1, 0x20);
  fmc176_spi_write(FMC176_AD9517, 0x01e0, 0x00);
  fmc176_spi_write(FMC176_AD9517, 0x01e1, 0x02);
  fmc176_spi_write(FMC176_AD9517, 0x0232, 0x01);
  fmc176_spi_write(FMC176_AD9517, 0x0018, 0x06);
  fmc176_spi_write(FMC176_AD9517, 0x0232, 0x01);
  delay_ms(1);

  fmc176_spi_write(FMC176_AD9517, 0x0018, 0x07);
  fmc176_spi_write(FMC176_AD9517, 0x0232, 0x01);
  delay_ms(1);

  fmc176_spi_verify(FMC176_AD9517, 0x0000, 0x18);
  fmc176_spi_verify(FMC176_AD9517, 0x0001, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0003, 0x51);
  fmc176_spi_verify(FMC176_AD9517, 0x0004, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0010, 0x7c);
  fmc176_spi_verify(FMC176_AD9517, 0x0011, 0x01);
  fmc176_spi_verify(FMC176_AD9517, 0x0012, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0013, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0014, 0x05);
  fmc176_spi_verify(FMC176_AD9517, 0x0015, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0016, 0x05);
  fmc176_spi_verify(FMC176_AD9517, 0x0017, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0018, 0x07);
  fmc176_spi_verify(FMC176_AD9517, 0x0019, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x001a, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x001b, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x001c, 0x02);
  fmc176_spi_verify(FMC176_AD9517, 0x001d, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x001e, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a0, 0x01);
  fmc176_spi_verify(FMC176_AD9517, 0x00a1, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a2, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a3, 0x01);
  fmc176_spi_verify(FMC176_AD9517, 0x00a4, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a5, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a6, 0x01);
  fmc176_spi_verify(FMC176_AD9517, 0x00a7, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a8, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00a9, 0x01);
  fmc176_spi_verify(FMC176_AD9517, 0x00aa, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00ab, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x00f0, 0x08);
  fmc176_spi_verify(FMC176_AD9517, 0x00f1, 0x08);
  fmc176_spi_verify(FMC176_AD9517, 0x00f4, 0x08);
  fmc176_spi_verify(FMC176_AD9517, 0x00f5, 0x08);
  fmc176_spi_verify(FMC176_AD9517, 0x0140, 0x42);
  fmc176_spi_verify(FMC176_AD9517, 0x0141, 0x42);
  fmc176_spi_verify(FMC176_AD9517, 0x0142, 0x42);
  fmc176_spi_verify(FMC176_AD9517, 0x0143, 0x43);
  fmc176_spi_verify(FMC176_AD9517, 0x0190, 0x21);
  fmc176_spi_verify(FMC176_AD9517, 0x0191, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0192, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0196, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0197, 0x80);
  fmc176_spi_verify(FMC176_AD9517, 0x0198, 0x02);
  fmc176_spi_verify(FMC176_AD9517, 0x0199, 0x21);
  fmc176_spi_verify(FMC176_AD9517, 0x019a, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x019b, 0x11);
  fmc176_spi_verify(FMC176_AD9517, 0x019c, 0x20);
  fmc176_spi_verify(FMC176_AD9517, 0x019d, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x019e, 0x21);
  fmc176_spi_verify(FMC176_AD9517, 0x019f, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x01a0, 0x11);
  fmc176_spi_verify(FMC176_AD9517, 0x01a1, 0x20);
  fmc176_spi_verify(FMC176_AD9517, 0x01a2, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x01e0, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x01e1, 0x02);
  fmc176_spi_verify(FMC176_AD9517, 0x0230, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0231, 0x00);
  fmc176_spi_verify(FMC176_AD9517, 0x0232, 0x00);

  if ((fmc176_spi_read(FMC176_AD9517, 0x001f, 0x1) & 0x1) == 0x0) {
    xil_printf("AD9517, PLL errors!!\n\r");
    return(1);
  }

  return(0);
}

// ***************************************************************************
// ***************************************************************************

void adc_capture(u32 sel, u32 qwcnt, u32 sa) {
  u32 dma_baseaddr;
  u32 cf_baseaddr;
  dma_baseaddr = (sel == 0) ? DMA0_BASEADDR : DMA1_BASEADDR;
  cf_baseaddr = (sel == 0) ? CF_ADC0_BASEADDR : CF_ADC1_BASEADDR;
  Xil_Out32((dma_baseaddr + 0x030), 0); // clear dma operations
  Xil_Out32((dma_baseaddr + 0x030), 1); // enable dma operations
  Xil_Out32((dma_baseaddr + 0x048), sa); // capture start address
  Xil_Out32((dma_baseaddr + 0x058), (qwcnt * 8)); // number of bytes
  Xil_Out32((cf_baseaddr + 0x00c), 0x0); // capture disable
  Xil_Out32((cf_baseaddr + 0x010), 0xf); // clear status
  Xil_Out32((cf_baseaddr + 0x014), 0xf); // clear status
  Xil_Out32((cf_baseaddr + 0x00c), (0x10000 | (qwcnt - 1)));
  do {delay_ms(1);}
  while ((Xil_In32(cf_baseaddr + 0x010) & 0x1) == 1);
  if ((Xil_In32(cf_baseaddr + 0x010) & 0x02) == 0x02) {
    xil_printf("overflow occured, capture may be corrupted\n\r");
  }
  microblaze_flush_dcache();
  microblaze_invalidate_dcache();
}

void adc_test(u32 sel, u32 mode, u32 format) {
  u32 i;
  u32 n;
  u32 rdata_2;
  u32 rdata[2];
  u32 edata[2];
  u32 cf_baseaddr;
  edata[0] = 0;
  edata[1] = 0;
  fmc176_spi_write(ad9250_sel(sel), 0x14, format);
  fmc176_spi_write(ad9250_sel(sel), 0x0d, mode);
  fmc176_spi_write(ad9250_sel(sel), 0xff, 0x01);
  adc_capture(sel, 16, DDR_BASEADDR);
  xil_printf("adc_test: sel(%d), mode(%2d), format(%2d)\n\r", sel, mode, format);
  cf_baseaddr = (sel == 0) ? CF_ADC0_BASEADDR : CF_ADC1_BASEADDR;
  if ((mode == 0x5) || (mode == 0x6)) {
    if (format == 0x1) return;
    Xil_Out32((cf_baseaddr+0x24), ((mode == 0x5) ? 0x03 : 0x00));
    delay_ms(10);
    Xil_Out32((cf_baseaddr+0x14), 0xffff);
    delay_ms(100);
    if ((Xil_In32(cf_baseaddr+0x14) & 0xffff) != 0) {
      xil_printf("  ERROR: PN status(%04x).\n\r", Xil_In32(cf_baseaddr+0x14));
    }
    return;
  }
  for (n = 0; n < 32; n++) {
    rdata_2 = Xil_In32(DDR_BASEADDR+(n*4));
    rdata[0] = (rdata_2>> 0) & 0xffff;
    rdata[1] = (rdata_2>>16) & 0xffff;
    for (i = 0; i < 2; i++) {
      if ((mode == 0x1) && (format == 0)) edata[i] = 0x2000;
      if ((mode == 0x2) && (format == 0)) edata[i] = 0x3fff;
      if ((mode == 0x3) && (format == 0)) edata[i] = 0x0000;
      if ((mode == 0x1) && (format == 1)) edata[i] = 0x0000;
      if ((mode == 0x2) && (format == 1)) edata[i] = 0x1fff;
      if ((mode == 0x3) && (format == 1)) edata[i] = 0x2000;
      if (((mode == 0x4) || (mode == 0x7)) && (n == 0)) edata[i] = (rdata[i] & 0xffff);
      if ((mode == 0x4) && (n > 0)) edata[i] = (edata[i] == 0x2aaa) ? 0x1555 : 0x2aaa;
      if ((mode == 0x7) && (n > 0)) edata[i] = (edata[i] == 0x3fff) ? 0x0000 : 0x3fff;
      if (rdata[i] != edata[i]) {
        xil_printf("  ERROR[%2d]: sel(%d), ch(%d), rcv(%08x), exp(%08x)\n\r", n, sel, i, rdata[i], edata[i]);
      }
    }
  }
}

u32 ad9250_setup() {

  u32 sel;
  u32 mode;

  Xil_Out32((JESD_BASEADDR + 0x14), 0x30); // reset
  delay_ms(100);

  Xil_Out32((JESD_BASEADDR + 0x04), 0x000c); // lanesync and scramble
  Xil_Out32((JESD_BASEADDR + 0x0c), 0x1f01); // frm and byte counts

  for (sel = 0; sel < 2; sel++) {
    fmc176_spi_write(ad9250_sel(sel), 0x5f, (0x16 | 0x1)); // trail bits, ilas normal & pd
    fmc176_spi_write(ad9250_sel(sel), 0x5e, 0x22); // m=2, l=2
    fmc176_spi_write(ad9250_sel(sel), 0x66, ((sel*2)+0)); // lane id
    fmc176_spi_write(ad9250_sel(sel), 0x67, ((sel*2)+1)); // lane id
    fmc176_spi_write(ad9250_sel(sel), 0x6e, 0x81); // scr, 2-lane
    fmc176_spi_write(ad9250_sel(sel), 0x70, 0x1f); // no. of frames per multi frame
    fmc176_spi_write(ad9250_sel(sel), 0x3a, 0x1e); // sysref enabled
    fmc176_spi_write(ad9250_sel(sel), 0x5f, (0x16 | 0x0)); // enable
    fmc176_spi_write(ad9250_sel(sel), 0x14, 0x00); // offset binary
    fmc176_spi_write(ad9250_sel(sel), 0x0d, 0x00); // test patterns
    fmc176_spi_write(ad9250_sel(sel), 0xff, 0x01);
    fmc176_spi_write(ad9250_sel(sel), 0xff, 0x00);
    fmc176_spi_verify(ad9250_sel(sel), 0x66, ((sel*2)+0)); // lane id
    fmc176_spi_verify(ad9250_sel(sel), 0x67, ((sel*2)+1)); // lane id
    fmc176_spi_verify(ad9250_sel(sel), 0x6e, 0x81); // scr, 2-lane
    fmc176_spi_verify(ad9250_sel(sel), 0x70, 0x1f); // no. of frames per multi frame
    fmc176_spi_verify(ad9250_sel(sel), 0x3a, 0x1e); // sysref ctrl
    fmc176_spi_verify(ad9250_sel(sel), 0x5f, 0x16); // enable
    fmc176_spi_verify(ad9250_sel(sel), 0x14, 0x00); // offset binary
    fmc176_spi_verify(ad9250_sel(sel), 0x0d, 0x00); // test patterns
    fmc176_spi_verify(ad9250_sel(sel), 0x6f, 0x01); // bytes per frame
    if (fmc176_spi_read(ad9250_sel(sel), 0x0a, 0x1) != 0x81) {
      xil_printf("AD9250_%d, PLL/link errors!!\n\r", sel);
    }
  }

  if (Xil_In32((JESD_BASEADDR + 0x08)) == 0x0) {
    xil_printf("JESD link/lane errors!!\n\r");
    return(0);
  }

  for (sel = 0; sel < 2; sel++) {
    for (mode = 1; mode <= 7; mode++) {
      adc_test(sel, mode, 0x0);
      adc_test(sel, mode, 0x1);
    }
    fmc176_spi_write(ad9250_sel(sel), 0x14, 0x00);
    fmc176_spi_write(ad9250_sel(sel), 0x0d, 0x00);
    fmc176_spi_write(ad9250_sel(sel), 0xff, 0x01);
  }

  return(0);
}

// ***************************************************************************
// ***************************************************************************
// clocks must be in mhz

u32 ad9129_dds_count(u32 sel, u32 sin_clk, u32 dac_clk) {
  if (sin_clk > (dac_clk/2)) {
    xil_printf("sin(%d) over NS rate(%d)\n\r", sin_clk, dac_clk);
  }
  xil_printf("dac_setup: sel(%d), dds(%d MHz)\n\r", sel, sin_clk);
  return((65536 * sin_clk)/dac_clk);
}

u32 ad9129_setup(u32 sel) {
  fmc176_spi_write(ad9129_sel(sel), 0x000, 0xa5); // soft reset
  fmc176_spi_write(ad9129_sel(sel), 0x000, 0x81); // soft reset
  fmc176_spi_write(ad9129_sel(sel), 0x030, 0x5c); // enable cross++, duty cycle correction off
  fmc176_spi_write(ad9129_sel(sel), 0x001, 0x68); // disable bias
  fmc176_spi_write(ad9129_sel(sel), 0x034, 0x5d); // normal mode (no interpolation)
  fmc176_spi_write(ad9129_sel(sel), 0x001, 0x48); // enable bias
  fmc176_spi_write(ad9129_sel(sel), 0x033, 0x13); // pll phase step = 1
  fmc176_spi_write(ad9129_sel(sel), 0x033, 0xd8); // pfd, clear pll lost bit
  fmc176_spi_write(ad9129_sel(sel), 0x033, 0xd0); // enable pll lost bit
  fmc176_spi_write(ad9129_sel(sel), 0x00c, 0x63); // enable dco
  fmc176_spi_write(ad9129_sel(sel), 0x00a, 0x40); // dll phase = 0, enable duty cycle correction
  fmc176_spi_write(ad9129_sel(sel), 0x00d, 0x06); // duty cycle correction bw low
  fmc176_spi_write(ad9129_sel(sel), 0x00a, 0xc0); // enable dll
  fmc176_spi_write(ad9129_sel(sel), 0x011, 0x01); // enable fifo phase reporting
  fmc176_spi_write(ad9129_sel(sel), 0x018, 0x40); // unsigned, interpolator off
  fmc176_spi_write(ad9129_sel(sel), 0x020, 0xc6); // full scale current = 33 mA
  fmc176_spi_write(ad9129_sel(sel), 0x021, 0x03); // fsc setting enable
  fmc176_spi_write(ad9129_sel(sel), 0x030, 0x46); // enable duty cycle correction
  fmc176_spi_write(ad9129_sel(sel), 0x011, 0x80); // fifo reset
  fmc176_spi_write(ad9129_sel(sel), 0x011, 0x80); // fifo reset
  fmc176_spi_write(ad9129_sel(sel), 0x001, 0x00); // enable output (iref)
  fmc176_spi_write(ad9129_sel(sel), 0x05c, 0x80); // enable parity
  fmc176_spi_write(ad9129_sel(sel), 0x007, 0x01); // frame pin mode is parity

  fmc176_spi_verify(ad9129_sel(sel), 0x034, 0x5d); // normal mode (no interpolation)
  fmc176_spi_verify(ad9129_sel(sel), 0x033, 0xd0); // enable pll lost bit
  fmc176_spi_verify(ad9129_sel(sel), 0x00c, 0x63); // enable dco
  fmc176_spi_verify(ad9129_sel(sel), 0x00d, 0x06); // duty cycle correction bw low
  fmc176_spi_verify(ad9129_sel(sel), 0x00a, 0xc0); // enable dll
  fmc176_spi_verify(ad9129_sel(sel), 0x018, 0x40); // unsigned, interpolator off
  fmc176_spi_verify(ad9129_sel(sel), 0x020, 0xc6); // full scale current = 33 mA
  fmc176_spi_verify(ad9129_sel(sel), 0x021, 0x03); // fsc setting enable
  fmc176_spi_verify(ad9129_sel(sel), 0x030, 0x46); // enable duty cycle correction
  fmc176_spi_verify(ad9129_sel(sel), 0x011, 0xc0); // fifo reset w ack
  fmc176_spi_verify(ad9129_sel(sel), 0x001, 0x00); // enable output (iref)
  fmc176_spi_verify(ad9129_sel(sel), 0x05c, 0x80); // enable parity
  fmc176_spi_verify(ad9129_sel(sel), 0x007, 0x21); // frame pin mode is parity
  delay_ms(100);
  return(0);
}

u32 ad9129_dds(u32 sel) {
  u32 cf_dac_baseaddr;
  cf_dac_baseaddr = (sel == 0) ? CF_DAC0_BASEADDR : CF_DAC1_BASEADDR;
  Xil_Out32((cf_dac_baseaddr + 0x04), 0x00000);
  Xil_Out32((cf_dac_baseaddr + 0x04), (0x10001 | ad9129_dds_count(sel, 98, 2458)));
  return(0);
}

u32 ad9129_clear(u32 sel) {
  delay_ms(10);
  fmc176_spi_write(ad9129_sel(sel), 0x00b, 0x80); // clear dll warnings
  fmc176_spi_write(ad9129_sel(sel), 0x00b, 0x00); // clear dll warnings
  fmc176_spi_write(ad9129_sel(sel), 0x05c, 0xa0); // enable parity/clear error
  fmc176_spi_write(ad9129_sel(sel), 0x05c, 0x80); // enable parity
  return(0);
}

u32 ad9129_monitor(u32 sel) {
  u32 status;
  status = 0;
  if ((fmc176_spi_read(ad9129_sel(sel), 0x035, 0x0) & 0xf0) != 0x80) {
    fmc176_spi_read(ad9129_sel(sel), 0x035, 0x1);
    xil_printf("AD9129_%d, PLL errors!!\n\r", sel);
    status = 1;
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x00e, 0x0) & 0xf1) != 0x81) {
    fmc176_spi_read(ad9129_sel(sel), 0x00e, 0x1);
    xil_printf("AD9129_%d, DLL errors!!\n\r", sel);
    status = 1;
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x05c, 0x0) & 0x03) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x05c, 0x1);
    xil_printf("AD9129_%d, parity errors!!\n\r", sel);
    status = 1;
  }
  if (fmc176_spi_read(ad9129_sel(sel), 0x05d, 0x0) != 0x0) {
    fmc176_spi_read(ad9129_sel(sel), 0x05d, 0x1);
    xil_printf("AD9129_%d, non zero rising edge parity errors!!\n\r", sel);
    status = 1;
  }
  if (fmc176_spi_read(ad9129_sel(sel), 0x05e, 0x0) != 0x0) {
    fmc176_spi_read(ad9129_sel(sel), 0x05e, 0x1);
    xil_printf("AD9129_%d, non zero falling edge parity errors!!\n\r", sel);
    status = 1;
  }
  return(status);
}

u32 ad9129_sed_1(u32 sel, u32 s1, u32 s0) {

  u32 status;
  u32 cf_dac_baseaddr;

  cf_dac_baseaddr = (sel == 0) ? CF_DAC0_BASEADDR : CF_DAC1_BASEADDR;

  xil_printf("dac_test: sel(%d), SED-1(0x%08x) SED-0(0x%08x)\n\r", sel, s1, s0);
  fmc176_spi_write(ad9129_sel(sel), 0x51, ((s0>> 0) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x52, ((s0>> 8) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x53, ((s0>>16) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x54, ((s0>>24) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x55, ((s1>> 0) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x56, ((s1>> 8) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x57, ((s1>>16) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x58, ((s1>>24) & 0xff));
  fmc176_spi_write(ad9129_sel(sel), 0x50, 0x00);
  fmc176_spi_write(ad9129_sel(sel), 0x50, 0x80);
  Xil_Out32((cf_dac_baseaddr + 0x40), s0);
  Xil_Out32((cf_dac_baseaddr + 0x44), s1);
  Xil_Out32((cf_dac_baseaddr + 0x04), 0x000000);
  Xil_Out32((cf_dac_baseaddr + 0x04), 0x110000);

  status = 0;
  delay_ms(10);
  fmc176_spi_write(ad9129_sel(sel), 0x50, 0xc0);
  fmc176_spi_write(ad9129_sel(sel), 0x50, 0x80);
  delay_ms(100);
  if ((fmc176_spi_read(ad9129_sel(sel), 0x50, 0) & 0x01) == 0x01) {
    fmc176_spi_read(ad9129_sel(sel), 0x50, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x50) SED FAIL!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x51, 0) & 0xff) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x51, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x51) SED S0[7:0]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x52, 0) & 0x3f) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x52, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x52) SED S0[13:8]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x53, 0) & 0xff) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x53, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x53) SED S1[7:0]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x54, 0) & 0x3f) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x54, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x54) SED S1[13:8]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x55, 0) & 0xff) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x55, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x55) SED S2[7:0]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x56, 0) & 0x3f) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x56, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x56) SED S2[13:8]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x57, 0) & 0xff) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x57, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x57) SED S3[7:0]!\n\r", sel);
  }
  if ((fmc176_spi_read(ad9129_sel(sel), 0x58, 0) & 0x3f) != 0x00) {
    fmc176_spi_read(ad9129_sel(sel), 0x58, 1);
    xil_printf("ERROR: AD9129_%d, Addr(0x58) SED S3[13:8]!\n\r", sel);
  }
  return(status);
}

void ad9129_sed(sel) {
  ad9129_sed_1(sel, 0x00000000, 0x0000aaaa);
  ad9129_sed_1(sel, 0x00000000, 0xaaaa0000);
  ad9129_sed_1(sel, 0x0000aaaa, 0x00000000);
  ad9129_sed_1(sel, 0xaaaa0000, 0x00000000);
  ad9129_sed_1(sel, 0x00000000, 0x00005555);
  ad9129_sed_1(sel, 0x00000000, 0x55550000);
  ad9129_sed_1(sel, 0x00005555, 0x00000000);
  ad9129_sed_1(sel, 0x55550000, 0x00000000);
  ad9129_sed_1(sel, 0x00000000, 0x00000000);
  ad9129_sed_1(sel, 0x55555555, 0xaaaaaaaa);
  ad9129_sed_1(sel, 0xaaaaaaaa, 0x55555555);
  return;
}

// ***************************************************************************
// ***************************************************************************

void iic_select(u32 sel) {
  u32 rdata;
  Xil_Out32((IIC_BASEADDR + 0x124), 0x000); // reset iic mux
  Xil_Out32((IIC_BASEADDR + 0x124), 0x001); // reset iic mux
  Xil_Out32((IIC_BASEADDR + 0x100), 0x002); // reset tx fifo
  Xil_Out32((IIC_BASEADDR + 0x100), 0x001); // enable iic
  Xil_Out32((IIC_BASEADDR + 0x108), 0x1e8); // mux select
  Xil_Out32((IIC_BASEADDR + 0x108), (0x200 | sel)); // enable selected device
  while ((Xil_In32(IIC_BASEADDR + 0x104) & 0x80) == 0x00) {delay_ms(1);}
  delay_ms(10);
  Xil_Out32((IIC_BASEADDR + 0x100), 0x002); // reset tx fifo
  Xil_Out32((IIC_BASEADDR + 0x100), 0x001); // enable iic
  Xil_Out32((IIC_BASEADDR + 0x108), 0x1e9); // mux select
  Xil_Out32((IIC_BASEADDR + 0x108), 0x201); // read back data
  while ((Xil_In32(IIC_BASEADDR + 0x104) & 0x40) == 0x40) {delay_ms(1);}
  rdata = Xil_In32(IIC_BASEADDR + 0x10c) & 0xff;
  xil_printf("iic_select: %02x\n\r", rdata);
}

u32 iic_read(u32 addr, u32 display) {
  u32 rdata;
  Xil_Out32((IIC_BASEADDR + 0x100), 0x002); // reset tx fifo
  Xil_Out32((IIC_BASEADDR + 0x100), 0x001); // enable iic
  Xil_Out32((IIC_BASEADDR + 0x108), (0x100 | 0x72)); // adv7511 select
  Xil_Out32((IIC_BASEADDR + 0x108), addr); // adv7511 address
  Xil_Out32((IIC_BASEADDR + 0x108), (0x101 | 0x72)); // adv7511 select
  Xil_Out32((IIC_BASEADDR + 0x108), 0x201); // adv7511 data
  while ((Xil_In32(IIC_BASEADDR + 0x104) & 0x40) == 0x40) {delay_ms(1);}
  delay_ms(10);
  rdata = Xil_In32(IIC_BASEADDR + 0x10c) & 0xff;
  if (display == 1) {
    xil_printf("iic_read: addr(%02x) data(%02x)\n\r", addr, rdata);
  }
  return(rdata);
}

void iic_write(u32 addr, u32 data) {
  Xil_Out32((IIC_BASEADDR + 0x100), 0x002); // reset tx fifo
  Xil_Out32((IIC_BASEADDR + 0x100), 0x001); // enable iic
  Xil_Out32((IIC_BASEADDR + 0x108), (0x100 | 0x72)); // adv7511 select
  Xil_Out32((IIC_BASEADDR + 0x108), addr); // adv7511 address
  Xil_Out32((IIC_BASEADDR + 0x108), (0x200 | data)); // adv7511 data
  while ((Xil_In32(IIC_BASEADDR + 0x104) & 0x80) == 0x00) {delay_ms(1);}
  delay_ms(10);
}

// ***************************************************************************
// ***************************************************************************

void ddr_video_wr(u32 m) {

  u32 n;
  u32 p;
  u32 dcnt;

  dcnt = 0;

  xil_printf("ddr write started\n\r");
  for (n = 0; n < V_ACTIVE(m); n++) {
    for (p = 0; p < H_ACTIVE(m); p++) {
      Xil_Out32((VIDEO_BASEADDR + (n*H_STRIDE(m)*4) + (p*4)), 0xFF);
      dcnt = dcnt + 1;
    }
  }
  xil_printf("ddr write completed (total %d)\n\r", dcnt);
  microblaze_flush_dcache();
  microblaze_invalidate_dcache();
}

// ***************************************************************************
// ***************************************************************************

void hdmi_setup(u32 m) {

  u32 data;

  delay_ms(1);
  data = Xil_In32(CF_CGEN_BASEADDR + (0x1f*4));
  if ((data & 0x1) == 0x0) {
    xil_printf("pixel clock (148.5MHz) out of lock (0x%04x)\n\r", data);
    return;
  }

  if (m == 1) {
    Xil_Out32((CF_CGEN_BASEADDR + (0x01*4)), 0x0000);
    Xil_Out32((CF_CGEN_BASEADDR + (0x02*4)), 0x01c7);
    Xil_Out32((CF_CGEN_BASEADDR + (0x03*4)), 0x0000);
    Xil_Out32((CF_CGEN_BASEADDR + (0x04*4)), 0x2083);
    Xil_Out32((CF_CGEN_BASEADDR + (0x05*4)), 0x034d);
    Xil_Out32((CF_CGEN_BASEADDR + (0x06*4)), 0x0000);
    Xil_Out32((CF_CGEN_BASEADDR + (0x07*4)), 0x0177);
    Xil_Out32((CF_CGEN_BASEADDR + (0x08*4)), 0x7c01);
    Xil_Out32((CF_CGEN_BASEADDR + (0x09*4)), 0x7fe9);
    Xil_Out32((CF_CGEN_BASEADDR + (0x0a*4)), 0x9000);
    Xil_Out32((CF_CGEN_BASEADDR + (0x0b*4)), 0x0800);
    Xil_Out32((CF_CGEN_BASEADDR + (0x01*4)), 0x0001);
    delay_ms(1);
    data = Xil_In32(CF_CGEN_BASEADDR + (0x1f*4));
    if ((data & 0x1) == 0x0) {
      xil_printf("pixel clock (74.25MHz) out of lock (0x%04x)\n\r", data);
      return;
    }
  }

  ddr_video_wr(m);
  iic_select(0x20);

  Xil_Out32((VDMA2_BASEADDR + 0x000), 0x4); // reset
  Xil_Out32((VDMA2_BASEADDR + 0x000), 0x0); // reset
  Xil_Out32((VDMA2_BASEADDR + 0x000), 0x3); // enable circular mode
  Xil_Out32((VDMA2_BASEADDR + 0x05c), VIDEO_BASEADDR); // start address
  Xil_Out32((VDMA2_BASEADDR + 0x060), VIDEO_BASEADDR); // start address
  Xil_Out32((VDMA2_BASEADDR + 0x064), VIDEO_BASEADDR); // start address
  Xil_Out32((VDMA2_BASEADDR + 0x058), (H_STRIDE(m)*4)); // h offset (2048 * 4) bytes
  Xil_Out32((VDMA2_BASEADDR + 0x054), (H_ACTIVE(m)*4)); // h size (1920 * 4) bytes
  Xil_Out32((VDMA2_BASEADDR + 0x050), V_ACTIVE(m)); // v size (1080)

  Xil_Out32((CF_HDMI_BASEADDR + 0x04), 0x0); // tpg & csc-bypass & enable
  Xil_Out32((CF_HDMI_BASEADDR + 0x08), ((H_WIDTH(m) << 16) | H_COUNT(m)));
  Xil_Out32((CF_HDMI_BASEADDR + 0x0c), ((H_DE_MIN(m) << 16) | H_DE_MAX(m)));
  Xil_Out32((CF_HDMI_BASEADDR + 0x10), ((V_WIDTH(m) << 16) | V_COUNT(m)));
  Xil_Out32((CF_HDMI_BASEADDR + 0x14), ((V_DE_MIN(m) << 16) | V_DE_MAX(m)));
  Xil_Out32((CF_HDMI_BASEADDR + 0x04), 0x1); // tpg & csc-bypass & enable

  // wait for hpd
  while ((iic_read(0x96, 0x01) & 0x80) != 0x80) {
    delay_ms(1);
  }

  iic_write(0x01, 0x00);
  iic_write(0x02, 0x18);
  iic_write(0x03, 0x00);
  iic_write(0x15, 0x01);
  iic_write(0x16, 0xb5);
  iic_write(0x18, 0x46);
  iic_write(0x40, 0x80);
  iic_write(0x41, 0x10);
  iic_write(0x48, 0x08);
  iic_write(0x49, 0xa8);
  iic_write(0x4c, 0x00);
  iic_write(0x55, 0x20);
  iic_write(0x56, 0x08);
  iic_write(0x96, 0x20);
  iic_write(0x98, 0x03);
  iic_write(0x99, 0x02);
  iic_write(0x9a, 0xe0);
  iic_write(0x9c, 0x30);
  iic_write(0x9d, 0x61);
  iic_write(0xa2, 0xa4);
  iic_write(0xa3, 0xa4);
  iic_write(0xa5, 0x44);
  iic_write(0xab, 0x40);
  iic_write(0xaf, 0x06);
  iic_write(0xba, 0x00);
  iic_write(0xd0, 0x3c);
  iic_write(0xd1, 0xff);
  iic_write(0xde, 0x9c);
  iic_write(0xe0, 0xd0);
  iic_write(0xe4, 0x60);
  iic_write(0xf9, 0x00);
  iic_write(0xfa, 0x00);
  iic_write(0x17, 0x02);
  iic_write(0x0a, 0x10);
  iic_write(0x0b, 0x8e);
  iic_write(0x0c, 0x00);
  iic_write(0x73, 0x01);
  iic_write(0x14, 0x02);

  iic_read(0x42, 0x01);
  iic_read(0xc8, 0x01);
  iic_read(0x9e, 0x01);
  iic_read(0x96, 0x01);
  iic_read(0x3e, 0x01);
  iic_read(0x3d, 0x01);
  iic_read(0x3c, 0x01);

  xil_printf("hdmi setup complete.\n\r");
  return;
}

u32 cnt2rgb(u32 d) {
  if (d > 0x00ffff) return(0xff0000);
  if (d > 0x0000ff) return(0x00ff00);
  if (d > 0x000000) return(0x0000ff);
  return(0);
}

void eyescan_setup(u32 m, u32 x, u32 y, u32 s) {

  u32 v, h, d;
  u32 sa0, da1, da2, da3, da4;

  Xil_Out32((JESD_BASEADDR + (0x15*4)), (VIDEO_BASEADDR + (H_ACTIVE(m)*y*4))); // start address
  Xil_Out32((JESD_BASEADDR + (0x16*4)), H_ACTIVE(m)); // hsize
  Xil_Out32((JESD_BASEADDR + (0x17*4)), (((x+0x101) << 16) | x)); // hmax, hmin
  Xil_Out32((JESD_BASEADDR + (0x11*4)), 0x0); // prescale
  Xil_Out32((JESD_BASEADDR + (0x12*4)), 0x017f81); // v-offset
  Xil_Out32((JESD_BASEADDR + (0x13*4)), 0x00400fc0); // h-offset
  Xil_Out32((JESD_BASEADDR + (0x14*4)), 0x0001); // h-offset
  Xil_Out32((JESD_BASEADDR + (0x10*4)), ((s == 1) ? 0x2 : 0x0)); // start
  Xil_Out32((JESD_BASEADDR + (0x10*4)), ((s == 1) ? 0x3 : 0x1)); // start
  while ((Xil_In32(JESD_BASEADDR + (0x18*4)) & 0x1) != 0x0) {delay_ms(1);}

  for (v = 0; v < 255; v++) {
    sa0 = VIDEO_BASEADDR + (H_ACTIVE(m)*y*4) + (H_ACTIVE(m)*v*4) + ((x+0x000)*4);
    da1 = VIDEO_BASEADDR + (H_ACTIVE(m)*y*4) + (H_ACTIVE(m)*v*4) + ((x+0x200)*4);
    da2 = VIDEO_BASEADDR + (H_ACTIVE(m)*y*4) + (H_ACTIVE(m)*v*4) + ((x+0x300)*4);
    da3 = VIDEO_BASEADDR + (H_ACTIVE(m)*y*4) + (H_ACTIVE(m)*v*4) + ((x+0x400)*4);
    da4 = VIDEO_BASEADDR + (H_ACTIVE(m)*y*4) + (H_ACTIVE(m)*v*4) + ((x+0x500)*4);
    for (h = 0; h < 129; h++) {
      if (s == 1) {
        d = Xil_In32(sa0);
        Xil_Out32(da1, cnt2rgb((d>> 0) & 0xffff));
        Xil_Out32(da3, cnt2rgb((d>>16) & 0xffff));
        sa0 = sa0 + 0x4;
        da1 = da1 + 0x4;
        da3 = da3 + 0x4;
        d = Xil_In32(sa0);
        Xil_Out32(da2, cnt2rgb((d>> 0) & 0xffff));
        Xil_Out32(da4, cnt2rgb((d>>16) & 0xffff));
        sa0 = sa0 + 0x4;
        da2 = da2 + 0x4;
        da4 = da4 + 0x4;
      } else {
        Xil_Out32(da1, cnt2rgb(Xil_In32(sa0)));
        sa0 = sa0 + 0x4;
        da1 = da1 + 0x4;
        Xil_Out32(da2, cnt2rgb(Xil_In32(sa0)));
        sa0 = sa0 + 0x4;
        da2 = da2 + 0x4;
      }
    }
  }
}

// ***************************************************************************
// ***************************************************************************

int main() {

  u8 n;
  u8 no_of_dac_channels;
  u8 no_of_adc_channels;

  init_platform();

  xil_printf("Enter number of dac channels (0, 1 or 2).\n\r");
  no_of_dac_channels = inbyte();
  no_of_dac_channels = ((no_of_dac_channels >= 48) && (no_of_dac_channels <= 50)) ?
    (no_of_dac_channels - 48) : 0;

  xil_printf("Enter number of adc channels (0, 1 or 2).\n\r");
  no_of_adc_channels = inbyte();
  no_of_adc_channels = ((no_of_adc_channels == 48) || (no_of_adc_channels == 50)) ?
    (no_of_adc_channels - 48) : 0;

  lcd_display(no_of_dac_channels, no_of_adc_channels);
  hdmi_setup(0);

  fmc176_spi_read(FMC176_AD9517,    0x03, 0x1);

  if (no_of_adc_channels >= 1) fmc176_spi_read(FMC176_AD9250_0,  0x01, 0x1);
  if (no_of_adc_channels >= 2) fmc176_spi_read(FMC176_AD9250_1,  0x01, 0x1);
  if (no_of_dac_channels >= 1) fmc176_spi_read(FMC176_AD9129_0,  0x7f, 0x1);
  if (no_of_dac_channels >= 2) fmc176_spi_read(FMC176_AD9129_1,  0x7f, 0x1);

  ad9517_setup();

  if (no_of_adc_channels >= 2) ad9250_setup();
  if (no_of_dac_channels >= 1) ad9129_setup(0);
  if (no_of_dac_channels >= 2) ad9129_setup(1);
  if (no_of_dac_channels >= 1) ad9129_sed(0);
  if (no_of_dac_channels >= 2) ad9129_sed(1);
  if (no_of_dac_channels >= 1) ad9129_dds(0);
  if (no_of_dac_channels >= 2) ad9129_dds(1);
  if (no_of_dac_channels >= 1) ad9129_clear(0);
  if (no_of_dac_channels >= 2) ad9129_clear(1);

  for (n = 0; n < 5; n++) {
    if (no_of_dac_channels >= 1) ad9129_monitor(0);
    if (no_of_dac_channels >= 2) ad9129_monitor(1);
    delay_ms(100);
  }

  xil_printf("running eye scan..\n\r");
  eyescan_setup(0, 0x50, 0x050, 0);
  eyescan_setup(0, 0x50, 0x250, 1);

  xil_printf("done\n\r");
  cleanup_platform();
  return(0);
}


// ***************************************************************************
// ***************************************************************************

